

# [ bb_sql.py ]


import thief_grr
from thief_grr.http import getIP
import sqlobject as SO
import cPickle
from os import getenv, stat, remove
from bb_conf import rc
from bb_error import BeeBBError
import bb_mail
from bb_image import fetchImage, isPhoto, isAvatar
from bb_valid import *




connection = SO.connectionForURI(rc['dbconnection'])
SO.sqlhub.processConnection = connection




################################
# Session implementation


class Session(SO.SQLObject):
	"""
	All sessions and their cookies are kept here.
	Cookies are stored as strings, actually pickled dictionaries.
	"""

	# columns
	bbsessid = SO.StringCol(notNone=True, alternateID=True, length=32)
	ip = SO.StringCol(default=lambda:getenv('REMOTE_ADDR'), length=24)
	lastused = SO.FloatCol(notNone=True, default=thief_grr.now())
	cookies = SO.StringCol(default=thief_grr.pickledDict)

	# foreign keys
	user = SO.ForeignKey("User")

	# M-M relations
	posts = SO.RelatedJoin('Post')	# Read for the first time

	################
	# functions

	def terminate(self):
		"""Terminates the session."""
		# change user's lastvisit
		if self.user.isreal:
			self.user.lastvisit = max(self.user.lastvisit, self.lastused)
		# clear all "posts" entries from other sessions of the same user
		# TODO: Use SQL DELETE to improve performance
		for session in set(self.user.sessions) - set([self]):
			for post in session.posts:
				session.removePost(post)
		# terminate session
		self.destroySelf()






################################
# User management


class TrialEmail(SO.SQLObject):
	"""
	This is where a new email address stays until it is confirmed as valid.
	"""

	# columns
	key = SO.StringCol(notNone=True, length=32, alternateID=True)
	oldemail = SO.StringCol(notNone=True, length=128, alternateID=True)
	newemail = SO.StringCol(notNone=True, length=128, alternateID=True)
	ip = SO.StringCol(default=lambda:getenv('REMOTE_ADDR'), length=24)
	created = SO.FloatCol(default=thief_grr.now())

	################
	# functions

	def ok(self):
		"""Makes the change in the User's table"""
		User.byEmail(self.oldemail).email = self.newemail
		self.destroySelf()

	def cancel(self, reason="Unknown."):
		"""Trial Email declined!  (Usually timed out.)"""
		body = '''
Your request to change your email address for %s failed miserably.
Reason: %s
''' % (rc['title'], reason)
		bb_mail.mail(to_addr=self.newemail, subject="Email change failed", body=body)
		self.destroySelf()


class TrialUser(SO.SQLObject):
	"""
	This is where a user stays until they submit the random string emailed to them.
	"""

	# columns
	key = SO.StringCol(notNone=True, length=32, alternateID=True)
	username = SO.UnicodeCol(notNone=True, length=24, alternateID=True)
	passwd = SO.StringCol(notNone=True, length=32)
	email = SO.StringCol(notNone=True, length=128, alternateID=True)
	ip = SO.StringCol(default=lambda:getenv('REMOTE_ADDR'), length=24)
	created = SO.FloatCol(default=thief_grr.now())

	################
	# functions

	def ok(self):
		"""Transfers a user from TrialUser to User."""
		User(username=self.username, passwd=self.passwd, email=self.email)
		self.destroySelf()

	def cancel(self, reason="Unknown."):
		"""Trial User declined!  (Usually timed out.)"""
		body = '''
Your request to register at %s failed miserably.
Reason: %s
''' % (rc['title'], reason)
		bb_mail.mail(to_addr=self.email, subject="Registration failed", body=body)
		self.destroySelf()


class TrialReset(SO.SQLObject):
	"""
	This is where a password reset request stays until they submit the
	random string emailed to them.
	"""

	# columns
	key = SO.StringCol(notNone=True, length=32, alternateID=True)
	passwd = SO.StringCol(notNone=True, length=32)
	ip = SO.StringCol(default=lambda:getenv('REMOTE_ADDR'), length=24)
	created = SO.FloatCol(default=thief_grr.now())

	# foreign keys
	user = SO.ForeignKey("User")

	################
	# functions

	def ok(self):
		"""Resets user's password."""
		self.user.passwd = thief_grr.MD5(self.passwd)
		self.destroySelf()

	def cancel(self, reason="Unknown."):
		"""Trial Reset declined!  (Usually timed out.)"""
		body = '''
Your request to reset your password at %s failed miserably.
Reason: %s
''' % (rc['title'], reason)
		bb_mail.mail(to_addr=self.user.email, subject="Password reset failed", body=body)
		self.destroySelf()


class User(SO.SQLObject):
	"""
	Registered users
	"""

	class sqlmeta:
		table = "users"

	# columns
	username = SO.UnicodeCol(notNone=True, length=24, alternateID=True)
	passwd = SO.StringCol(notNone=True, length=32)
	level = SO.FloatCol(default=0.1, notNone=True)
	email = SO.StringCol(notNone=True, length=128, alternateID=True)
	created = SO.FloatCol(default=thief_grr.now())
	lastvisit = SO.FloatCol(default=None)
	checkemail = SO.FloatCol(default=0.0)
	memolist = SO.StringCol(default=cPickle.dumps([]))
	errorlist = SO.StringCol(default=cPickle.dumps([]))
	optiondict = SO.StringCol(default=thief_grr.pickledDict)

	# 1-1 relations
	details = SO.SingleJoin('UserDetails', joinColumn='user_id')
	# 1-M relations
	threads = SO.MultipleJoin('Thread', joinColumn='author_id', orderBy='created')
	posts = SO.MultipleJoin('Post', joinColumn='author_id', orderBy='created')
	sessions = SO.MultipleJoin('Session', joinColumn='user_id', orderBy='-lastused')
	trialresets = SO.MultipleJoin('TrialReset', joinColumn='user_id')

	# M-M relations
	boards = SO.RelatedJoin('Board')

	################
	# functions

	def hasPasswd(self, passwd):
		return thief_grr.MD5(passwd) == self.passwd

	def addError(self, error):
		errors = self.getErrors()
		errors.append(unicode(error))
		self.errorlist = cPickle.dumps(errors)
	
	def rmError(self, error):
		errors = self.getErrors()
		errors.remove(error)
		self.errorlist = cPickle.dumps(errors)

	def addMemo(self, memo):
		memos = self.getMemos()
		memos.append(unicode(memo))
		self.memolist = cPickle.dumps(memos)
	
	def rmMemo(self, memo):
		memos = self.getMemos()
		memos.remove(memo)
		self.memolist = cPickle.dumps(memos)

	def setOption(self, option, value):
		options = self.getOptions()
		options[option] = value
		self.optiondict = cPickle.dumps(options)
	
	def rmOption(self, option):
		options = self.getOptions()
		try:
			options.pop(option)
			self.optiondict = cPickle.dumps(options)
		except KeyError:
			pass
	
	def resetPasswd(self, passwd):
		if not self.hasPasswd(passwd):
			validatePasswd(passwd)
			key = thief_grr.randstr()
			TrialReset(key=key, user=self, passwd=passwd)
			bb_mail.TrialReset(to_addr=self.email, key=key)
	
	def changeEmail(self, email):
		validateEmail(email)
		key = thief_grr.randstr()
		TrialEmail(old=self.email, new=email, key=key, ip=getIP())
		bb_mail.TrialEmail(to_addr=email, key=key, user=self)

	def usePhoto(self, photostr):
		try:
			if photostr:
				photo, msg = fetchImage(photostr, directory='photos', filename=self.username, max_size=thief_grr.str2bytes(rc['max_photo_size']))
				isPhoto(photo)
				photo = self.username
			else:
				photo, msg = '', 'No photo.'
			self.photo = photo
			self.addMemo(msg)
		except Exception, e:
			self.addError(str(e))
			self.photo = ''
		return self.photo

	def useAvatar(self, avatarstr):
		#TODO: Check the file before downloading
		#TODO: Use old avatar if the new one sucks
		try:
			if avatarstr:
				avatar, msg = fetchImage(avatarstr, directory='avatars', filename=self.username, max_size=thief_grr.str2bytes(rc['max_avatar_size']))
				isAvatar(avatar)
				avatar = self.username
			else:
				avatar, msg = '', 'No avatar.'
			#self.avatar = avatar
			self.addMemo(msg)
		except Exception, e:
			self.addError(str(e))
			#self.avatar = ''
		#return self.avatar
		return avatar

	################
	# permissions

	def canReadBoard(self, board):
		return self.level >= board.rlevel

	def canReadThread(self, thread):
		return self.canReadBoard(thread.board)

	def canReadPost(self, post):
		return self.canReadThread(post.thread)

	def canPostOnThread(self, thread):
		return self.canPostOnBoard(thread.board) and (thread.locked == False)

	def canPostOnBoard(self, board):
		if self.level >= 1:
			return True
		elif self.level >= board.wlevel and not board.islocked():
			return True
		elif board in self.allboards:
			return True
		else:
			return False
	
	def canCreateOnBoard(self, board):
		if self.level >= 1:
			return True
		elif board in self.allboards:
			return True
		else:
			return False

	def canAdministrateBoard(self, board):
		if self.level >= 1:
			return True
		elif board.parent:
			return self.canCreateOnBoard(board.parent)
		else:
			return False
	
	def canDeleteBoard(self, board):
		return self.canAdministrateBoard(board)

	def canLockBoard(self, board):
		return self.canCreateOnBoard(board)

	def canDeleteThread(self, thread):
		return self.canCreateOnBoard(thread.board)

	def canLockThread(self, thread):
		# add a "or (self is thread.author)" if you want to allow
		# thread authors locking thir threads.  It's nasty, though.
		return self.canDeleteThread(thread)

	def canDeletePost(self, post):
		return self.canAdministrateBoard(post.thread.board)

	def canEditPost(self, post):
		return self.canDeletePost(post) or (self is post.author and post is post.thread.lastpost)

	def canMessage(self, user):
		return self.level >= 1 or (user.options['accept_messages'] and self.isreal)

	################
	# action functions

	def pm(self, to, message):
		if not self.canMessage(to):
			raise BeeBBError, "You are not permitted to send messages to %s." % to.username
		else:
			bb_mail.pm(sender=self, recepient=to, message=message)
	
	def createThread(self, board, head=None):
		if not self.canPostOnBoard(board=board):
			raise BeeBBError, "User %s cannot create a new thread on board %s." % (self.username, board.title)
		elif not head:
			raise BeeBBError, "You must specify a head for your post"
		else:
			return Thread(authorID=self.id, boardID=board.id)
	
	def createPost(self, thread, body, head=None):
		if not self.canPostOnThread(thread=thread):
			raise BeeBBError, "User %s cannot post on this thread" % self.username
		elif not head:
			raise BeeBBError, "You must specify a head for your post"
		else:
			return Post(authorID=self.id, threadID=thread.id, head=head, body=body)
	
	def createBoard(self, parent, title, descr='', rlevel=0.0, wlevel=0.1):
		if not self.canCreateOnBoard(parent):
			raise BeeBBError, "User %s cannot create a sub-board under this board." % self.username
		else:
			return Board(parentID=parent.id, title=title, descr=descr, rlevel=rlevel, wlevel=wlevel)
	
	def deleteThread(self, thread):
		if not self.canDeleteThread(thread):
			raise BeeBBError, "Permission denied.  Ha!"
		else:
			subject = thread.subject
			thread.rm()
			return "Thread <em>%s</em> deleted succesfully." % subject

	def deletePost(self, post):
		if not self.canDeletePost(post):
			raise BeeBBError, "Permission denied.  Ha!"
		else:
			post.rm()
			return "Post deleted succesfully."

	def deleteBoard(self, board):
		if not self.canDeleteBoard(board):
			raise BeeBBError, "Permission denied.  Ha!"
		else:
			board.rm()
			return "Board <em>%s</em> deleted succesfully." % board.title

	def deleteObject(self, object):
		if object._table == 'thread':
			return self.deleteThread(object)
		elif object._table == 'post':
			return self.deletePost(object)
		elif object._table == 'board':
			return self.deleteBoard(object)
		else:
			raise BeeBBError, "Error trying to delete something.  What?"

	################
	# properties

	def getFullname(self):
		if self.details:
			return ' '.join([self.details.firstname, self.details.lastname])
		else:
			return ''
	
	def getErrors(self):
		return cPickle.loads(self.errorlist)

	def getMemos(self):
		return cPickle.loads(self.memolist)

	def getOptions(self):
		return cPickle.loads(self.optiondict)

	def getAllboards(self):
		ab = self.boards
		for board in self.boards:
			ab += board.allchildren
		return ab

	def getIsonline(self):
		if not self.sessions:
			return False
		else:
			return thief_grr.now() - self.sessions[0].lastused <= thief_grr.str2secs(rc['max_user_idle'])
	
	def getIsvirtual(self):
		return self in virtualusers

	def getIsreal(self):
		return not self.isvirtual

	fullname = property(getFullname)
	errors = property(getErrors)
	memos = property(getMemos)
	options = property(lambda self: thief_grr.easydict(self.getOptions(), dd=rc, default=''))
	allboards = property(getAllboards)
	created_human = property(lambda self: thief_grr.unix2human(self.created))
	lastvisit_human = property(lambda self: thief_grr.unix2human(self.lastvisit, nonemsg='Who knows?'))
	isonline = property(getIsonline)
	isvirtual = property(getIsvirtual)
	isreal = property(getIsreal)


class UserDetails(SO.SQLObject):
	"""All non-techy details"""
	# columns
	firstname = SO.UnicodeCol(default='', length=32)
	lastname = SO.UnicodeCol(default='', length=48)
	sex = SO.EnumCol(default='', enumValues=['', 'female', 'male'])
	birthday = SO.FloatCol(default=None)
	photo = SO.StringCol(default='', length=24)
	avatar = SO.StringCol(default='', length=24)
	location = SO.UnicodeCol(default='', length=64)
	timeoffset = SO.IntCol(default=0, notNone=True)
	about = SO.UnicodeCol(default='', length=512)
	website = SO.StringCol(default='', length=160)
	publicemail = SO.StringCol(default='', length=128)
	signature = SO.UnicodeCol(default='', length=128)
	updated = SO.FloatCol(default=None)
	# foreign keys
	user = SO.ForeignKey("User")








################################
# bb


class Board(SO.SQLObject):
	"""
	All boards and (...sub)boards are kept here.
	"""

	# columns
	title = SO.UnicodeCol(notNone=True, length=48)
	descr = SO.UnicodeCol(default=None, length=256)
	prec = SO.FloatCol(default=0.0)
	locked = SO.BoolCol(default=False, notNone=True)
	rlevel = SO.FloatCol(default=0.0)
	wlevel = SO.FloatCol(default=0.1)

	# foreign keys
	parent = SO.ForeignKey("Board")

	# 1-M relations
	threads = SO.MultipleJoin('Thread', joinColumn='board_id', orderBy='-updated')
	children = SO.MultipleJoin('Board', joinColumn='parent_id', orderBy='prec')

	# M-M relations
	moderators = SO.RelatedJoin('User')

	################
	# functions

	def turnkey(self):
		self.locked = not self.locked
		if self.locked:
			state = "locked"
		else:
			state = "unlocked"
		return 'Board <a href="index.py?board=%d">%s</a> has been %s' % (self.id, self.title, state)
	
	def islocked(self):
		if self.locked:
			return True
		elif self.parent:
			return self.parent.islocked()
		else:
			return False
	
	def rm(self):
		# remove all threads from all subboards
		for thread in self.allthreads:
			thread.rm()
		# remove all subboards
		for board in self.allchildren:
			board.destroySelf()
		# remove this board
		self.destroySelf()
		return "%s and everything under it, got removed.  You bastard..." % self.title

	################
	# properties

	def getAdmins(self):
		"""
		Returns the admins of this board, i.e., the moderators of its parent.
		"""
		if self.parent:
			return self.parent.moderators + self.parent.admins
		else:
			return list(User.selectBy(level=1.0))

	def getAllchildren(self):
		"""
		This returns all the subboards under this one.
		"""
		ac = self.children
		for child in self.children:
			ac += child.allchildren
		return ac

	def getAllthreads(self):
		"""
		This returns not only this board's threads, but also all the threads
		of it's (...sub)boards as well.
		"""
		at = self.threads
		for child in self.children:
			for thread in child.allthreads:
				at.append(thread)
		return at

	def getSubthreads(self):
		"""
		This returns the threads of this board's children (and not grand-children...).
		"""
		st = []
		for child in self.children:
			for subthread in child.threads:
				st.append(subthread)
		return st

	def getAllposts(self):
		"""
		This returns not only posts in this board's threads, but also all
		the posts of the threads in it's (...sub)boards as well.
		"""
		ap = []
		for thread in self.allthreads:
			for post in thread.posts:
				ap.append(post)
		return ap

	def getLastpost(self):
		"""
		This should return the latest post, among those posted on this
		board or on any of the boards "under" it.  We don't care about
		last edited.  Only the creation of each post matters.
		"""
		lp = None
		for post in self.allposts:
			if (not lp) or post.created > lp.created:
				lp = post
		return lp

	def getUpdated(self):
		try:
			return max(self.lastpost.created, self.lastpost.updated)
		except:
			return None

	def getSiblings(self):
		"""
		Returns a list of all siblings of this board.
		"""
		if not self.parent:
			return []
		else:
			return self.parent.children
	
	def getPath(self):
		"""
		Returns a list of the "path", starting from the root board.
		"""
		p = []
		t = self
		while t != rootboard and t.parent:
			p.append(t)
			t = t.parent
		p.append(t)
		p.reverse()
		return p
		

	admins = property(getAdmins)
	allthreads = property(getAllthreads)
	subthreads = property(getSubthreads)
	allposts = property(getAllposts)
	lastpost = property(getLastpost)
	updated = property(getUpdated)
	siblings = property(getSiblings)
	allchildren = property(getAllchildren)
	path = property(getPath)
	level = property(lambda self: len(self.path))
	author_id = property(lambda self: self.author.id)
	author_username = property(lambda self: self.author.username)


class Thread(SO.SQLObject):
	"""
	This is the "root" of a thread.  It has no body of it's own,
	and each time it is created, usually we create a Post too, serving as
	the thread's first body.
	"""

	# columns
	created = SO.FloatCol(default=thief_grr.now())
	views = SO.IntCol(default=0)
	locked = SO.BoolCol(default=False)

	# foreign keys
	author = SO.ForeignKey("User")
	board = SO.ForeignKey("Board")

	# 1-M relations
	posts = SO.MultipleJoin('Post', joinColumn='thread_id', orderBy='created')

	################
	# functions

	def turnkey(self):
		self.locked = not self.locked
		if self.locked:
			state = "locked"
		else:
			state = "unlocked"
		return 'Thread <a href="thread.py?thread=%d">%s</a> has been %s' % (self.id, self.subject, state)

	def rm(self):
		for post in self.posts:
			post.destroySelf()
		self.destroySelf()
		return "Thread %s and all posts in it got removed." % self.subject

	################
	# properties

	def getLastpost(self):
		lp = self.posts[-1:]
		try:
			return lp[0]
		except IndexError:
			return None

	def getUpdated(self):
		try:
			return max(self.lastpost.created, self.lastpost.updated)
		except:
			return None
	
	def getSubject(self):
		try:
			return unicode(self.posts[0].head)
		except:
			return "Empty thread."

	lastpost = property(getLastpost)
	updated = property(getUpdated)
	updated_human = property(lambda self: thief_grr.unix2human(self.updated))
	created_human = property(lambda self: thief_grr.unix2human(self.created))
	author_id = property(lambda self: self.author.id)
	subject = property(getSubject)


class Post(SO.SQLObject):
	"""
	Holds all posts, both thread's "roots" as well as replies.
	"""

	# columns
	head = SO.UnicodeCol(default=None, length=72)
	body = SO.UnicodeCol(notNone=True, length=4096)
	created = SO.FloatCol(default=thief_grr.now())
	updated = SO.FloatCol(default=None)

	# foreign keys
	author = SO.ForeignKey("User")
	thread = SO.ForeignKey("Thread")

	# M-M relations
	sessions = SO.RelatedJoin('Session')	# Sessions who read this post


	################
	# functions

	def rm(self):
		self.destroySelf()
		return "Post got removed."

	def isNewFor(self, user):
		"""
		Actually works correctly only if checked against a .lastpost, which is how
		it is checked within BeeBB.
		"""

		# virtual users have no read/unread posts
		if user.isvirtual:
			return False
		# is this post read by any of the user's sessions?
		for session in user.sessions:
			if self in session.posts:
				return False
		# was it posted before user's lastvisit?
		return (self.updated or self.created) > user.lastvisit

	################
	# properties

	def getHeader(self):
		if self.head:
			return self.head
		else:
			return self.thread.subject
	
	header = property(getHeader)
	created_human = property(lambda self: thief_grr.unix2human(self.created))
	updated_human = property(lambda self: thief_grr.unix2human(self.updated))
	subject = property(lambda self: self.head or self.thread.subject)
	board = property(lambda self: self.thread.board)
	author_id = property(lambda self: self.author.id)
	author_username = property(lambda self: self.author.username)







################################
# website content


class Web(object):
	"""This class contains the content-related SQLObject classes"""

	class Announcement(SO.SQLObject):
		head = SO.UnicodeCol(default=None)
		body = SO.UnicodeCol(notNone=True)





################################
# main / constants

try:
	rootboard = Board.get(1)
	root = User.get(1)
	guest = User.get(2)
	exuser = User.get(3)
	virtualusers = [guest, exuser]
	realusers = list(User.select())
	# get rid of virtual users
	for vu in virtualusers:
		realusers.remove(vu)

except:
	pass
